/*
 *                            COPYRIGHT
 *
 *  PCB, interactive printed circuit board design
 *  Copyright (C) 1994,1995 Thomas Nau
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Contact addresses for paper mail and Email:
 *  Thomas Nau, Schlehenweg 15, 88471 Baustetten, Germany
 *  Thomas.Nau@rz.uni-ulm.de
 *
 */

static	char	*rcsid = "$Header: /sda7/users/nau/src/pcb/pcb-1.4.0/RCS/print.c,v 1.4 1996/01/13 16:35:39 nau Exp $";

/* printing routines
 */
#include <math.h>
#include <time.h>

#include "global.h"

#include "data.h"
#include "dev_ps.h"
#include "file.h"
#include "error.h"
#include "misc.h"
#include "print.h"

/* ---------------------------------------------------------------------------
 * some local identifiers
 */
static	PrintDeviceTypePtr	Device;
static	PrintInitType		DeviceFlags;

/* ----------------------------------------------------------------------
 * queries color from X11 database and calls device  command
 * black is default on errors
 */
static void SetPrintColor(Pixel X11Color)
{
	XColor	rgb;
	int		result;

		/* query best matching color from X server */
	rgb.pixel = X11Color;
	result = XQueryColor(Dpy,
		DefaultColormapOfScreen(XtScreen(Output.Toplevel)),
		&rgb);

		/* create a PS command to set this color */
	if (result == BadValue && result == BadColor)
		rgb.red = rgb.green = rgb.blue = 0;
	Device->SetColor(rgb);
}

/* ----------------------------------------------------------------------
 * generates printout
 * mirroring, rotating and scaling were already determined by a dialog
 *
 * output is written either to a file or, if the command starts with
 * a pipe symbol, piped to a command (if supported by the device)
 * whitespaces have already been removed from 'Command'
 *
 * 'MirrorFlag' is negated because (0,0) is assumed to be in the lower/left
 * corner for printing devices (X11 is different)
 *
 * 'PackageOnly' selects a special mode where only packages, pins, pads
 * and vias are printed
 *
 * all offsets are in mil
 */
int Print(char *Command, float Scale,
	Boolean MirrorFlag, Boolean RotateFlag,
	Boolean ColorFlag, Boolean PackageOnly,
	PrintDeviceTypePtr PrintDevice, MediaTypePtr Media,
	Position OffsetX, Position OffsetY)
{
	char		*filename,
				*completeFilename,
				*message;
	Cardinal	i;
	Boolean		error;

		/* save pointer in global identifier */
	Device = PrintDevice;

		/* check if we have to open a pipe */
	if (*Command == '|')
	{
			/* does the driver support pipes (no seek()...) ? */
		if (!Device->HandlesPipes)
		{
			Message(False, "device '%s' can't handle piped input\n",
				Device->Name);
			return(1);
		}

			/* pipes only work with single files */
		if (Device->UsedFiles != 1)
		{
			Message(False, "wrong device driver configuration; only single"
				" files are supported in combination with pipes\n");
			return(1);
		}

			/* OK, open it */
		if ((DeviceFlags.FP[0] = popen(Command+1, "w")) == NULL)
		{
			PopenErrorMessage(Command+1);
			return(1);
		}
	}
	else
	{
			/* OK, one or more filepointers are required */
		if (Device->UsedFiles > MAX_PRINT_FILES || !Device->UsedFiles)
		{
			Message(False, "number of required filehandles is out of range"
				" [1,%i]\n", MAX_PRINT_FILES);
			return(1);
		}

			/* expand filename without additional directories */
		if ((filename = ExpandFilename(NULL, Command)) == NULL)
			filename = Command;
		for (i = 0; i < Device->UsedFiles; i++)
		{
			size_t	length;

				/* add suffix to filename
				 * Sander, thanks for the bug-fix (sander@centerline.com)
				 */
			length = strlen(filename) + 1 +
				(Device->Suffix[i] ? strlen(Device->Suffix[i]) : 0);
			completeFilename = MyCalloc(length, sizeof(char), "Print()");
			sprintf(completeFilename, "%s%s",
				filename, EMPTY(Device->Suffix[i]));

				/* try to open all the files, close already opened
				 * ones in case of an error
				 */
			DeviceFlags.FP[i] = CheckAndOpenFile(completeFilename);
			if (DeviceFlags.FP[i] == NULL)
			{
				OpenErrorMessage(completeFilename);
				SaveFree(completeFilename);
				while(i)
					fclose(DeviceFlags.FP[--i]);
				return(1);
			}
			SaveFree(completeFilename);
		}
	}

		/* 'MirrorFlag' has to be inverted because X11 and printing devices
		 * use different Y-axis; it has also to be inverted if we look to
		 * the bottom side
		 */
	if (!Settings.ShowSolderSide)
		MirrorFlag = !MirrorFlag;

		/* set the info struct for the device driver */
	DeviceFlags.SelectedMedia = Media;
	DeviceFlags.MirrorFlag = MirrorFlag;
	DeviceFlags.RotateFlag = RotateFlag;
	DeviceFlags.OffsetX = OffsetX;
	DeviceFlags.OffsetY = OffsetY;
	DeviceFlags.Scale = Scale;

		/* set bounding box coordinates; layout has already been checked
		 * to be non-empty
		 */
	DeviceFlags.BoundingBox = *GetDataBoundingBox(PCB->Data);

		/* initialize the driver and print error message if necessary */
	if ((message = Device->Init(&DeviceFlags)) != NULL)
	{
		Message(False, message);
		if (*Command == '|')
			pclose(DeviceFlags.FP[0]);
		else
			for (i = 0; i < Device->UsedFiles; i++)
				fclose(DeviceFlags.FP[i]);
		return(1);
	}

		/* pass drilling information */
	if (PCB->PinOn || PackageOnly)
		ALLPIN_LOOP(PCB->Data, Device->Drill(pin););
	if (PCB->ViaOn || PackageOnly)
		VIA_LOOP(PCB->Data, Device->Drill(via););

		/* print signal layers only if allowed;
		 * print data, vias and pins at last
		 */
	for (i = MAX_LAYER-1; !PackageOnly && i != -1; i--)
	{
		LayerTypePtr	layer = LAYER_ON_STACK(i);

		if (layer->On)
		{
			if (ColorFlag)
				SetPrintColor(layer->Color);
			Device->Layer(layer, GetLayerGroupNumber(layer));
		}
	}

	if (PCB->ElementOn || PackageOnly)
		ELEMENT_LOOP(PCB->Data,
			if ((TEST_FLAG(ONBOTTOMFLAG, element) != 0) == SWAP_IDENT ||
				PCB->InvisibleObjectsOn)
			{
				if (ColorFlag)
					SetPrintColor(PCB->ElementColor);
				Device->ElementPackage(element);
			}	
		);
	if (PCB->PinOn || PackageOnly)
		ELEMENT_LOOP(PCB->Data,
			if (ColorFlag)
				SetPrintColor(PCB->PinColor);
			PIN_LOOP(element, Device->PinOrVia(pin););
			PAD_LOOP(element,
				if ((TEST_FLAG(ONBOTTOMFLAG, pad) != 0) == SWAP_IDENT ||
					PCB->InvisibleObjectsOn)
					Device->Pad(pad);
			);
		);
	if (PCB->ViaOn || PackageOnly)
		{
			if (ColorFlag)
				SetPrintColor(PCB->ViaColor);
			VIA_LOOP(PCB->Data, Device->PinOrVia(via););
		}

	Device->Exit();

	if (*Command == '|')
		return(pclose(DeviceFlags.FP[0]));

	error = False;
	for (i = 0; i < Device->UsedFiles; i++)
		error |= fclose(DeviceFlags.FP[i]);
	return((int) error);
}
